import subprocess
import psutil # need to be added as a dependency


def is_sclang_running():
    running = False
    for process in psutil.process_iter():
        if 'sclang' in process.name():
            running = True
    return running

def boot_supercollider():
    if not is_sclang_running():
        subprocess.Popen(['sclang', '/home/elie/Bureau/Livecoding/tech_setup/start_foxdot.sc'])




from .SCFilesHandling import write_sc_renardo_files_in_user_config
from .SuperColliderInstance import SupercolliderInstance
from .PulsarInstance import PulsarInstance
from .RenardoTUI import RenardoTUI
from renardo_gatherer.samples_download import SPackManager

import argparse
import time

class RenardoApp:

    def __init__(self):
        self.sc_instance = None
        self.spack_manager = SPackManager()
        self.args = RenardoApp.parse_args()
        self.sc_instance = SupercolliderInstance()
        self.pulsar_instance = PulsarInstance()
        self.launch()

    def launch(self):

        # if args.cli:
        # if args.dir:
        #     try:
        #         # Use given directory
        #         FoxDotCode.use_sample_directory(args.dir)
        #     except OSError as e:
        #         # Exit with last error
        #         import sys, traceback
        #         sys.exit(traceback.print_exc(limit=1))

        # if args.startup:
        #     try:
        #         FoxDotCode.use_startup_file(args.startup)
        #     except OSError as e:
        #         import sys, traceback
        #         sys.exit(traceback.print_exc(limit=1))
        # if args.no_startup:
        #     FoxDotCode.no_startup()

        if self.args.create_scfiles:
            write_sc_renardo_files_in_user_config()

        if not (self.args.no_tui or self.args.pipe or self.args.foxdot_editor):
            RenardoTUI(self).run()

        if self.args.boot:
            print("Launching Renardo SC module with SCLang...")
            self.sc_instance.start_sclang_subprocess()
            output_line = self.sc_instance.read_stdout_line()
            while "Welcome to" not in output_line:
                print(output_line[:-1]) # remove \n at the end to avoid double newline
                output_line = self.sc_instance.read_stdout_line()
            self.sc_instance.evaluate_sclang_code("Renardo.start;")
            time.sleep(3)

        if self.args.pipe:
            from renardo_lib import handle_stdin, FoxDotCode
            # Just take commands from the CLI
            handle_stdin()
        elif self.args.foxdot_editor:
            from renardo_lib import FoxDotCode
            # Open the GUI
            from FoxDotEditor.Editor import workspace
            FoxDot = workspace(FoxDotCode).run()
        elif self.args.no_tui:
            print("You need to choose a launching mode : TUI, --pipe or --foxdot-editor...")
        print("Quitting...")
    
    @staticmethod
    def parse_args():
        parser = argparse.ArgumentParser(
            prog="renardo",
            description="Live coding with Python and SuperCollider",
            epilog="More information: https://renardo.org/"
        )
        parser.add_argument('-N', '--no-tui', action='store_true', help="does start renardo TUI")
        parser.add_argument('-p', '--pipe', action='store_true', help="run Renardo from the command line interface")
        parser.add_argument('-f', '--foxdot-editor', action='store_true', help="run Renardo with the classic FoxDot code editor")
        # parser.add_argument('-d', '--dir', action='store', help="use an alternate directory for looking up samples")
        # parser.add_argument('-s', '--startup', action='store', help="use an alternate startup file")
        # parser.add_argument('-n', '--no-startup', action='store_true', help="does not load startup.py on boot")
        # store_false => boot default value = True WTF
        parser.add_argument('-b', '--boot', action='store_true', help="Boot SuperCollider Renardo instance automatically")
        parser.add_argument('-c', '--create-scfiles', action='store_true', help="Create Renardo class file and startup file in SuperCollider user conf dir.")

        return parser.parse_args()




import os
import subprocess
from sys import platform
import time
import pathlib

import psutil

from renardo.SCFilesHandling import SC_USER_CONFIG_DIR


class SupercolliderInstance:

    def __init__(self):
        self.sclang_process = None

        if platform == "win32":
            sc_dir = next(pathlib.Path("C:\\Program Files").glob("SuperCollider*")) # to match SuperCollider-3.version like folders
            os.environ["PATH"] +=  f"{sc_dir};"
            sclang_path = sc_dir / "sclang.exe"
            #self.sclang_exec = [str(sclang_path), str(SC_USER_CONFIG_DIR / 'start_renardo.scd')]
            self.sclang_exec = [str(sclang_path), '-i', 'scqt']
        else:
            #self.sclang_exec = ["sclang",  str(SC_USER_CONFIG_DIR / 'start_renardo.scd')]
            self.sclang_exec = ["sclang", '-i', 'scqt']


    def start_sclang_subprocess(self):
        if not self.is_sclang_running():
            #print("Auto Launching Renardo SC module with SCLang...")
            self.sclang_process = subprocess.Popen(
                args=self.sclang_exec,
                #shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
            )

    def read_stdout_line(self):
        if self.sclang_process.returncode is None:
           return self.sclang_process.stdout.readline().decode("utf-8")

    def read_stderr_line(self):
        if self.sclang_process.returncode is None:
            return self.sclang_process.stderr.readline().decode("utf-8")

    def evaluate_sclang_code(self, code_string):
        raw = code_string.encode("utf-8") + b"\x1b"
        self.sclang_process.stdin.write(raw)
        self.sclang_process.stdin.flush()

        # TODO : find a way to consistently stop sclang and scsynth when renardo stops/dies
        # TODO : find a way to name/tag the sclang/synth processes with name renardo to find it better
        # TODO : Use name renardo for scsynth audio server (for example with JACK Driver)

    def __del__(self):
        pass
        # self.popen.kill() # TODO: fix that the destructor is not called
        # need to clarify the launch and close process of foxdot/renardo !
        # self.popen.wait()

    def is_sclang_running(self):
        running = False
        for process in psutil.process_iter():
            if 'sclang' in process.name():
                running = True
        return running



import os

from pathlib import Path
from sys import platform

SC_USER_CONFIG_DIR = None

# default config path
# on windows AppData/Roaming/renardo
# on Linux ~/.config/renardo
# on MacOS /Users/<username>/Library/Application Support/renardo
if platform == "linux" or platform == "linux2" :
    home_path = Path.home()
    SC_USER_CONFIG_DIR = home_path / '.local' / 'share' / 'SuperCollider'
elif platform == "darwin":
    home_path = Path.home()
    SC_USER_CONFIG_DIR = home_path / 'Library' / 'Application Support' / 'SuperCollider'
elif platform == "win32":
    appdata_local_path = Path(os.getenv('LOCALAPPDATA'))
    SC_USER_CONFIG_DIR = appdata_local_path / 'SuperCollider'

SC_USER_EXTENSIONS_DIR = SC_USER_CONFIG_DIR / 'Extensions'

SCLANG_PROCESS = None

def is_renardo_scfiles_installed():
    return (
        (SC_USER_EXTENSIONS_DIR / 'Renardo.sc').exists()
        and (SC_USER_EXTENSIONS_DIR / 'StageLimiter.sc').exists()
        and (SC_USER_CONFIG_DIR / 'start_renardo.scd').exists()
    )

def write_sc_renardo_files_in_user_config():
    renardo_sc_class = '''
        Renardo
        {
            classvar server;
            classvar midiout;

            *start
            { | remote = false |

                server = Server.default;
                server.options.memSize = 8192 * 16; // increase this if you get "alloc failed" messages
                server.options.maxNodes = 1024 * 32; // increase this if you are getting drop outs and the message "too many nodes"
                server.options.numOutputBusChannels = 2; // set this to your hardware output channel size, if necessary
                server.options.numInputBusChannels = 2; // set this to your hardware output channel size, if necessary

                if (remote, {
                    server.options.bindAddress = "0.0.0.0"; // allow connections from any address
                });

                server.boot();

                OSCFunc(
                    {
                        arg msg, time, addr, port;
                        var fn;
                        // Get local filename
                        fn = msg[1].asString;
                        // Print a message to the user
                        ("Loading SynthDef from" + fn).postln;
                        // Add SynthDef to file
                        fn = File(fn, "r");
                        fn.readAllString.interpret;
                        fn.close;
                    },
                    'foxdot'
                );

                StageLimiterBis.activate(2);
                "Listening for messages from Renardo".postln;
            }

            *startRemote
            {
                this.start(true);
            }

            *midi
            {
                arg port=0;
                MIDIClient.init;
                midiout = MIDIOut(port);
                OSCFunc(
                    {
                        arg msg, time, addr, port;
                        var note, vel, sus, channel, nudge;
                        // listen for specific MIDI trigger messages from FoxDot
                        note    = msg[2];
                        vel     = msg[3];
                        sus     = msg[4];
                        channel = msg[5];
                        nudge   = msg[6];
                        SystemClock.schedAbs(time + nudge, {midiout.noteOn(channel, note, vel)});
                        SystemClock.schedAbs(time + nudge + sus, {midiout.noteOff(channel, note, vel)});
                    },
                    'foxdot_midi'
                );
                ("Sending Renardo MIDI messages to" + MIDIClient.destinations[port].name).postln;
            }
        }
   '''

    stagelimiter_sc_class = '''
        // Batuhan Bozkurt 2009
        StageLimiterBis
        {
            classvar lmSynth, lmFunc, activeSynth;
            
            *activate
            { |numChannels = 2|
                fork
                {
                    lmFunc = 
                    { 
                        { 
                            activeSynth = 
                                Synth(\\stageLimiter,
                                    target: RootNode(Server.default), 
                                    addAction: \\addToTail
                                );
                        }.defer(0.01) 
                    };
                    lmSynth = SynthDef(\\stageLimiter, 
                    {
                        var input = In.ar(0, numChannels);
                        input = Select.ar(CheckBadValues.ar(input, 0, 0), [input, DC.ar(0), DC.ar(0), input]);
                        ReplaceOut.ar(0, Limiter.ar(input)) ;
                    }).add;
                    Server.default.sync;
                    lmFunc.value;
                    CmdPeriod.add(lmFunc);
                    "StageLimiter active".postln;
                }
            }
            
            *deactivate
            {
                activeSynth.free;
                CmdPeriod.remove(lmFunc);
                "StageLimiter inactive...".postln;
            }
        }
    '''

    renardo_start_code = '''
        Renardo.start;
        Renardo.midi;
    '''

    with open(SC_USER_EXTENSIONS_DIR / 'StageLimiter.sc', mode="w") as file:
        file.write(stagelimiter_sc_class)

    with open(SC_USER_EXTENSIONS_DIR / 'Renardo.sc', mode="w") as file:
        file.write(renardo_sc_class)

    with open(SC_USER_CONFIG_DIR / 'start_renardo.scd', mode="w") as file:
        file.write(renardo_start_code)





#!/usr/bin/env python

from setuptools import setup

with open("README.md", "r") as f:
    long_description=f.read()

setup(
    name='renardo',
    version="0.9.4.dev1",
    description='Launcher/config editor for Renardo livecoding environment',
    author='Elie Gavoty',
    author_email='eliegavoty@free.fr',
    license='cc-by-sa-4.0',
    url='http://renardo.org/',
    packages=[
        'renardo',
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    # entry_points={'gui_scripts' : ['FoxDotEditor = FoxDotEditor.__init__:main']},
    # data_files=[('', 'LICENSE')],
    package_data = {'renardo': ['RenardoTUI.tcss'],},
    install_requires=[
        'renardo-lib>=0.9.4.dev0',
        'FoxDotEditor>=0.9.4.dev0',
        'renardo_gatherer>=0.1.3.dev0',
        'psutil',
        'textual',
    ],
    entry_points={
        'console_scripts': [
            'renardo = renardo:entrypoint',
        ]
    }
)